--[[----

	DeclinerAllowed is the main player check function, the bouncer, which decides if a player is a stranger or not.
	Usage, such as in another addon or a WeakAura:

			DeclinerAllowed("name-realm")

		Returns true if you recently whispered "name", they're in your group, or online in your guild, friends, or bnet.
		Returns false if they are a complete stranger.

			DeclinerAllowed("name-realm","action")

		Returns true or false for "name" based on your settings for that action.
		Actions are "chan", "duel", "pet", "group", "guild", "chart", or "trade".

	For example, DeclinerAllowed("Guildie") will return true because Guildie is online in your guild.

	An example of the second one, let's say your options will decline duel requests from guild members and realm friends.
	Mcstabyrouge is a guildie, but not a friend, DeclinerAllowed("Mcstabyrouge","duel") will return false.
	You add them as a friend, DeclinerAllowed("Mcstabyrouge","duel") still returns false.

	----    If no realm is provided for "name", your realm is assumed.    ----

--------

	DeclinerTrack returns the amount of times you recently whispered players or they were declined.
	Players are cleared as Decliner forgets whispers or they don't spam you after some time.
	Usage, such as in another addon or a WeakAura:

			DeclinerTrack("action","name-realm")

		Returns the number of times that action occurred recently from "name".

			DeclinerTrack("action")

		Returns how many players are currently listed for that action.

	Actions are "whisper", "chan", "duel", "pet", "group", "guild", "chart", or "trade".
	"whisper" is the amount of times _you_ whispered "name" while the others are the times they were blocked for that action.

	For example, Suprawsom spammed a duel request 5 times, so DeclinerTrack("duel","Suprawsom") will return 5, then 4, and so on as time goes by.
	Once the last action is cleared, DeclinerTrack("duel","Suprawsom") will be nil, allowing Decliner to once again notify you of any further duels from them.
	Another example, you whispered Chanter 9 times due to a conversation, so DeclinerTrack("whisper","Chanter") will be 9.
	They will be authorized for whatever your options are set to for whispers, such as trading, until that number counts down to nil.

	An example of the second variable, let's say you are standing in front of Orgrimmar on the Area 52 server.
	14 players just spammed you with duel invites one right after another. DeclinerTrack("duel") will return 14.
	Depending on how many times each player spammed you, that number will count down erratically.
	It may jump to nil if all 14 only sent one request or hang around at a lower number for a while when the single requests are cleared.

	----    If no realm is provided for "name", your realm is assumed.    ----

----]]--

 -- localize lua functions

local date=date
local error=error
local format=string.format
local gmatch=string.gmatch
local gsub=string.gsub
local insert=table.insert
local ipairs=ipairs
local lower=string.lower
local match=string.match
local next=next
local random=math.random
local remove=table.remove
local select=select
local sort=table.sort
local time=time
local tonumber=tonumber
local tostring=tostring
local type=type
local upper=string.upper

 -- localize Blizzard functions

local After=C_Timer.After
local BNGetFriendGameAccountInfo=BNGetFriendGameAccountInfo
local BNGetNumFriendGameAccounts=BNGetNumFriendGameAccounts
local BNGetNumFriends=BNGetNumFriends
local CancelDuel=CancelDuel
local CancelPetDuel=C_PetBattles.CancelPVPDuel
local ClosePetition=ClosePetition
local CloseTrade=CloseTrade
local CreateFrame=CreateFrame
local ct=CopyTable
local DeclineChannelInvite=DeclineChannelInvite
local DeclineGroup=DeclineGroup
local DeclineGuild=DeclineGuild
local DisableDropDown=UIDropDownMenu_DisableDropDown
local EnableDropDown=UIDropDownMenu_EnableDropDown
local GetCVar=GetCVar
local GetFriendInfo=GetFriendInfo
local GetGuildRosterInfo=GetGuildRosterInfo
local GetNumFriends=GetNumFriends
local GetNumGroupMembers=GetNumGroupMembers
local GetNumGuildMembers=GetNumGuildMembers
local GetPetitionInfo=GetPetitionInfo
local HideUIPanel=HideUIPanel
local IsInGroup=IsInGroup
local IsInGuild=IsInGuild
local IsInRaid=IsInRaid
local PlaySoundKitID=PlaySoundKitID
local SendAddonMessage=SendAddonMessage
local SetAutoDeclineGuildInvites=SetAutoDeclineGuildInvites
local SetCVar=SetCVar
local split=strsplit
local StaticPopupSpecial_Hide=StaticPopupSpecial_Hide
local StaticPopup_Hide=StaticPopup_Hide
local StopSound=StopSound
local Ticker=C_Timer.NewTicker
local UnitName=UnitName
local wipe=wipe

 -- initialize

local _
local a,L=...
local d={} -- data
local e={} -- events
local f={} -- functions
local g='|cff00ff00'
local r='|cffff0000'
local s=ChatTypeInfo.SYSTEM
local DCL=CreateFrame('frame',a)
local DCLPanel=CreateFrame('frame',a..'Panel')
DCL:RegisterEvent('ADDON_LOADED')
DCL:RegisterEvent('PLAYER_LOGIN')
DCL:RegisterEvent('PLAYER_LOGOUT')
d.file={
	ch='Interface\\RaidFrame\\ReadyCheck-Ready',
	cr='Interface\\RaidFrame\\ReadyCheck-NotReady',
	ha='Interface\\Icons\\Spell_Misc_EmotionHappy',
	sa='Interface\\Icons\\Spell_Misc_EmotionSad',
}
d.def={
	enable={chan=true,duel=true,pet=true,group=true,guild=true,chart=true,trade=true,},
	tell={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	group={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	guild={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	friend={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	bnet={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	msg={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	chan={emote=true,say=true,yell=true,[1]=true,[2]=true,[26]=true,},
	gen={openinv=false,msgs=false,disable=false,},
	open={'1','invite','portal',},
}
local dlog={}
declinerlog={}
d.PLtrack={whisper={}}
for k in next,d.def.enable do d.PLtrack[k]={} end
d.ply,d.stp,d.bnet,d.friend,d.group,d.guild={},{},{},{},{},{}

 -- global string strip
 -- originally from Phanx

function f.pattern(txt)
	txt=gsub(txt,'%%%d?$?s','') -- removes %s
	txt=gsub(txt,':','') -- removes colons
	txt=gsub(txt,'%(','') -- removes (
	txt=gsub(txt,'%)','') -- removes )
	txt=gsub(txt,'%.','') -- removes periods
	txt=gsub(txt,'^%s+','') -- removes space from beginning of string
	txt=gsub(txt,'%s+$','') -- removes space from end of string
	return txt
end

 -- Blizzard global strings used
 -- everything in caps can be found in FrameXML/GlobalStrings.lua
 -- this table is only for keeping strings in one place

d.S={
	ADD=ADD.."/"..REMOVE,
	BAT=BATTLENET_OPTIONS_LABEL.." "..FRIENDS,
	CAL=gsub(BN_INLINE_TOAST_ALERT,BATTLENET_OPTIONS_LABEL,CHAT),
	CHA=CHANNEL,
	CHS=CHANNELS,
	DIO=DISABLE.." "..OPENING.." "..GROUP.." ",
	DIS=DISABLE.." ",
	DUE=DUEL,
	EMO=EMOTE,
	ENA=ENABLE.." "..DECLINE,
	GEN=GENERAL,
	GRM=GROUP.." "..MEMBERS,
	GRO=GROUP,
	GUI=GUILD,
	GUM=GUILD.." "..MEMBERS,
	INV=f.pattern(GUILDEVENT_TYPE_INVITE):gsub("^%l",upper),
	LOO=LFG_TITLE,
	PET=PET_BATTLE_PVP_QUEUE,
	PTI=f.pattern(PETITION_TITLE),
	REA=lower(READY),
	REF=f.pattern(FRIENDS_LIST_REALM).." "..FRIENDS,
	REQ=GUILDINFOTAB_APPLICANTS_NONE,
	SAY=SAY,
	TRA=TRADE,
	VER=lower(GAME_VERSION_LABEL),
	WHI=f.pattern(AUTOCOMPLETE_LABEL_INTERACTED),
	WOW=BNET_CLIENT_WOW,
	YEL=YELL,
}

 -- matching tables for event, sound, and text filters

d.EV={
	CHANNEL_INVITE_REQUEST=true,
	DUEL_REQUESTED=true,
	GUILD_INVITE_REQUEST=true,
	PARTY_INVITE_REQUEST=true,
	PETITION_SHOW=true,
	PET_BATTLE_PVP_DUEL_REQUESTED=true,
	TRADE_SHOW=true,
}
d.SO={
	igCharacterInfoClose=840,
	igCharacterInfoOpen=839,
	igMainMenuClose=851,
	igMainMenuOpen=850,
	igPlayerInvite=880,
	igQuestListClose=876,
	igQuestListOpen=875,
}
d.SY={
	duel=gsub(DUEL_REQUESTED,'%%%d?$?s','(.*)'),
	pet=gsub(PET_BATTLE_PVP_DUEL_REQUESTED,'%%%d?$?s','(.*)'),
	guild=gsub(gsub(ERR_INVITED_TO_GUILD_SSS,'%%%d?$?s','.*'),'|H.-|h.-|h','(.*)'),
	group=gsub(ERR_INVITED_TO_GROUP_SS,'|H.-|h.-|h','(.*)'),
	group1=gsub(ERR_INVITED_ALREADY_IN_GROUP_SS,'|H.-|h.-|h','(.*)'),
}
d.TR=gsub(ERR_INITIATE_TRADE_S,'%%%d?$?s','(.*)')
d.UI={[ERR_DUEL_CANCELLED]=true,[ERR_PETBATTLE_DECLINED]=true,[ERR_TRADE_CANCELLED]=true,}

 -- internal bouncer, what DeclinerAllowed uses

function f.allowed(name,option,request,sup,missingrealm,bypass)
	if type(name)~='string' then return true end
	name=match(name,'%-') and name or name.."-"..d.realm
	option=tostring(option) missingrealm=match(name,"(.*%-)")
	if d.def.enable[option] then
		bypass=false
		if d.var.gen.disable or not d.var.enable[option] then
			if not sup then f.debug(name..': allowed - ['..option..'] disabled') end
			d.block=false return true
		else if not sup then f.debug(name..': ['..option..'] enabled') end end
	else bypass=true if not sup then f.debug(name..': ['..option..'] invalid, bypassed') end end
	if bypass or not d.var.tell[option] then
		if d.PLtrack.whisper[name] or d.PLtrack.whisper[missingrealm] then
			if not sup then f.debug(name..': allowed - whispered') end
			d.block=false return true
		elseif #d.PLtrack.whisper<1 then if not sup then f.debug(name..': no one recently whispered') end
		else if not sup then f.debug(name..': did not recently whisper') end end
	else if not sup then f.debug(name..': tell disabled for '..option) end end
	if bypass or option~='group' then if bypass or not d.var.group[option] then
		if IsInGroup() then
			if d.group[name] or d.group[missingrealm] then
				if not sup then f.debug(name..': allowed - group member') end
				d.block=false return true
			else if not sup then f.debug(name..': not a group member') end end
		else if not sup then f.debug(name..': not in a group') end end
	else if not sup then f.debug(name..': group disabled for '..option) end end end
	if bypass or option~='guild' then if bypass or not d.var.guild[option] then
		if IsInGuild() then
			if d.guild[name] or d.guild[missingrealm] then
				if not sup then f.debug(name..': allowed - guild member') end
				d.block=false return true
			else if not sup then f.debug(name..': not a guild member') end end
		else if not sup then f.debug(name..': not in a guild') end end
	else if not sup then f.debug(name..': guild disabled for '..option) end end end
	if bypass or not d.var.friend[option] then
		if d.friend[name] or d.friend[missingrealm] then
			if not sup then f.debug(name..': allowed - realm friend') end
			d.block=false return true
		else if not sup then f.debug(name..': not a realm friend') end end
	else if not sup then f.debug(name..': friend disabled for '..option) end end
	if bypass or not d.var.bnet[option] then
		if d.bnet[name] or d.bnet[missingrealm] then
			if not sup then f.debug(name..': allowed - bnet friend') end
			d.block=false return true
		else if not sup then f.debug(name..': not a bnet friend') end end
	else if not sup then f.debug(name..': bnet disabled for '..option) end end
	if not sup then f.debug(name..': declined') end
	if not sup and not request then
		if option=='duel' or option=='pet' or option=='trade' then d.UIinfo=true end
		After(0.05,function() d.block=false end)
	end
	return false
end

 -- debug handler

function f.debug(msg,...)
	msg=match(msg,':') and '    '..msg or match(msg,'^%u') and msg..' '..date('%y%m%d%H%M%S',time()) or msg
	local args={n=select('#',...),...}
	local argstr
	if args[1] then for i=1,args.n do
		argstr=argstr and argstr..', '..tostring(args[i]) or ' - args: '..tostring(args[i])
		if i==5 then break end
	end else argstr='' end
	insert(dlog,msg..argstr)
	if #dlog>999 then remove(dlog,1) end
end

 -- allows all group invites after player last said certain words
 -- changing zones or joining a group will cancel this effect
 -- effect ends after 5 minutes since last trigger
 -- being in a group will not trigger this effect
 -- if a number is used as a trigger word, it only works by itself
 -- (numbers have too many false positives in normal conversation and item links)

function f.openinvite(msg,sender)
	if sender~=d.player or d.var.gen.openinv or IsInGroup() then return end
	for _,v in next,d.var.open do
		if type(tonumber(v))=='number' then
			if msg==v then
				f.timer('openinv') f.debug('OPENINV',v,'['..msg..']') return
			end
		else
			for m in gmatch(msg,'%w*'..v..'%w*') do if m==v then
				f.timer('openinv') f.debug('OPENINV',m,'['..msg..']') return
			end end
		end
	end
end

 -- chat output
 -- the idea for faking a yellow system message came from Jaliborc's Scrap
 -- they use ChatFrame_MessageEventHandler while I use AddMessage

function f.print(msg,action,name,data)
	if d.var.gen.disable or d.var.gen.msgs then return end
	if name==1 then DEFAULT_CHAT_FRAME:AddMessage(r..a..": |r"..msg,s.r,s.g,s.b) return end
	if action then
		name=match(name,'%-') and name or name.."-"..d.realm
		if f.timer('player',action,name)>1 then
		f.debug(name..': '..d.PLtrack[action][name]..' times recently') return end
		f.debug(name..': first recent occurrence')
		if d.var.msg[action] then return end
		name=match(name,"%-(.*)")==d.realm and match(name,"(.*)%-") or name
		msg=format(msg,"|Hplayer:"..name.."|h["..name.."]|h",data)
	end
	local notfound=true
	for i=1,NUM_CHAT_WINDOWS do if _G['ChatFrame'..i]:IsEventRegistered('CHAT_MSG_SYSTEM') then
		notfound=false _G['ChatFrame'..i]:AddMessage(r..a..":|r "..msg,s.r,s.g,s.b,s.id)
		if f.P then f.P:Prat_PostAddMessage('','',_G['ChatFrame'..i],'',r..a..":|r "..msg,s.r,s.g,s.b,s.id) end
	end end
	if notfound then
		DEFAULT_CHAT_FRAME:AddMessage(r..a..": |r"..msg,s.r,s.g,s.b,s.id)
		if f.P then f.P:Prat_PostAddMessage('','',DEFAULT_CHAT_FRAME,'',r..a..":|r "..msg,s.r,s.g,s.b,s.id) end
	end
end

 -- dummy function to expose f.allowed

function f.requestallowed(name,action,sup)
	if type(name)~='string' then
		error('Usage: DeclinerAllowed( character-realm[, action] )\n        See top of dclcore.lua for description',2)
	else
		if not sup then f.debug('REQUEST',name,action) end
		return f.allowed(name,action,true,sup)
	end
end

 -- dummy function to expose d.PLtrack

function f.requesttracked(action,name)
	if type(action)~='string' or (name and type(name)~='string') or not d.PLtrack[action] then
		error('Usage: DeclinerTrack( action[, character-realm] )\n        See top of dclcore.lua for description',2)
	elseif name then
		return d.PLtrack[action][(match(name,'%-') and name or name.."-"..d.realm)]
	else
		return #d.PLtrack[action]
	end
end

 -- timer to monitor globals and track openinvites/players

function f.timer(which,reason,name)
	if which=='global' then
		if reason then
			if not DeclinerAllowed then
				f.debug("--- "..(name or 'unknown addon').." deleted DeclinerAllowed ---")
				DeclinerAllowed=f.requestallowed
			end
			if not DeclinerTrack then
				f.debug("--- "..(name or 'unknown addon').." deleted DeclinerTrack ---")
				DeclinerTrack=f.requesttracked
			end
		else
			DeclinerAllowed=f.requestallowed
			DeclinerTrack=f.requesttracked
		end
		if not name then After(1,function() f.timer('global',true) end) end
	elseif which=='openinv' then
		d.OItrack=d.OItrack and d.OItrack+1 or 1
		d.OpenInv=d.OItrack>0
		After(300,function()
			d.OItrack=d.OItrack>0 and d.OItrack-1 or 0
			d.OpenInv=d.OpenInv and d.OItrack>0
		end)
	elseif which=='player' then
		name=match(name,'%-') and name or name.."-"..d.realm
		if not d.PLtrack[reason][name] then
			d.PLtrack[reason][name]=1
			f.timer('playerloop',reason,name)
		else
			d.PLtrack[reason][name]=d.PLtrack[reason][name]+1
		end
		return d.PLtrack[reason][name]
	elseif which=='playerloop' then
		After(60,function()
			if d.PLtrack[reason][name]==1 then
				d.PLtrack[reason][name]=nil
			else
				d.PLtrack[reason][name]=d.PLtrack[reason][name]-1
				f.timer('playerloop',reason,name)
			end
		end)
	end
end

 -- event functions

function e.PLAYER_LOGIN()
	GuildRoster()
	ShowFriends()
	if Decliner_SV then d.var=ct(Decliner_SV)
	for k1,v1 in next,d.def do for k2,v2 in next,v1 do if d.var[k1]==nil then d.var[k1]=d.def[k1]
	elseif type(d.var[k1][k2])~=type(v2) and k1~='open' then d.var[k1][k2]=d.def[k1][k2] end end end
	for k1,v in next,d.var do if d.def[k1]==nil then d.var[k1]=nil elseif type(v)=='table' then
	for k2 in next,v do if d.def[k1][k2]==nil and k1~='open' then d.var[k1][k2]=nil end end end end
	else d.var=ct(d.def) end
	d.ch=L.ch d.pf=L.pf
	d.player,d.realm=UnitFullName('player')
	d.player=d.player.."-"..d.realm
	f.panel() f.register()
	DCL:RegisterEvent('BN_FRIEND_INFO_CHANGED')
	DCL:RegisterEvent('CHAT_MSG_ADDON')
	DCL:RegisterEvent('CHAT_MSG_WHISPER_INFORM')
	DCL:RegisterEvent('FRIENDLIST_UPDATE')
	DCL:RegisterEvent('GROUP_ROSTER_UPDATE')
	DCL:RegisterEvent('GUILD_ROSTER_UPDATE')
	DCL:UnregisterEvent('PLAYER_LOGIN')
	RegisterAddonMessagePrefix('DCLDBG')
	hooksecurefunc('PlaySound',f.silence)
	ChatFrame_AddMessageEventFilter('CHAT_MSG_SYSTEM',f.system)
	if Prat then local module=Prat.Addon:GetModule('History',true) if module then f.P=module end end
	f.print(d.S.VER.." "..GetAddOnMetadata(a,'Version').." "..d.S.REA,nil,1)
end

function e.PLAYER_LOGOUT()
	Decliner_SV=ct(d.var)
end

function e.ADDON_LOADED(addon)
	if addon==a then f.timer('global') dlog=declinerlog
	else f.timer('global',true,addon) end
end

function e.BN_FRIEND_INFO_CHANGED(id)
	if id then
		for j=1,(BNGetNumFriendGameAccounts(id)) do
			local _,name,client,realm=BNGetFriendGameAccountInfo(id,j)
			if client==d.S.WOW and name then d.bnet[name.."-"..(realm or "")]=true end
		end
	elseif d.nbnet~=select(2,BNGetNumFriends()) then
		local count
		count,d.nbnet=BNGetNumFriends()
		wipe(d.bnet)
		for i=1,count do for j=1,(BNGetNumFriendGameAccounts(i)) do
			local _,name,client,realm=BNGetFriendGameAccountInfo(i,j)
			if client==d.S.WOW and name then d.bnet[name.."-"..(realm or "")]=true end
		end end
	end
end

function e.CHANNEL_INVITE_REQUEST(...)
	f.debug('CHAN',...)
	local chan,name=...
	if not f.allowed(name,'chan') then
		StaticPopup_Hide('CHAT_CHANNEL_INVITE')
		DeclineChannelInvite()
		f.print(L.DECLINED_CHAN,'chan',name,chan)
	end
end

--[[ begin CHAT_MSG_ADDON

	transparency note: Due to scammers using certain scripts, I
	feel the need to explain the following piece of code. For
	debugging purposes, this CHAT_MSG_ADDON usage allows me to
	"download" a user's log ingame if I can whisper them. My
	standard debugging method is for a user to upload their
	Decliner.lua in a ticket on CurseForge, however this code
	makes it extremely convenient to view that log. So far, I
	have only used it with a willing tester on working versions
	of Decliner, but it may be useful in the future for a
	guildie, another friend, or anyone that I may meet. This can
	only give me what is shown in declinerlog and nothing else.
	You can view the first 30 lines using /dump declinerlog
	or all of it in the Decliner.lua saved file found at
	WTF\Account\<account name or id>\SavedVariables\Decliner.lua

]]

function e.CHAT_MSG_ADDON(prefix,msg,channel,sender)
	if not (prefix==d.pf and channel==d.ch and match(msg,'^'..a..'%d+:%d+$')) then return end
	msg=gsub(msg,a,'')
	local msg1,msg2=split(':',msg)
	msg1=tonumber(msg1)>#dlog and #dlog or tonumber(msg1)
	msg2=tonumber(msg2)>#dlog and #dlog or tonumber(msg2)
	local lines=msg2-msg1+1
	SendAddonMessage(prefix,lines..':'..#dlog,channel,sender)
	if #dlog<1 then return
	elseif ChatThrottleLib then
		for i=msg1,msg2 do
			ChatThrottleLib:SendAddonMessage('BULK',prefix,dlog[i],channel,sender)
		end
	else
		Ticker(0.05,function()
			SendAddonMessage(prefix,dlog[msg1],channel,sender)
			msg1=msg1+1
		end,lines)
	end
end

--[[ end CHAT_MSG_ADDON ]]

function e.CHAT_MSG_CHANNEL(msg,sender,_,_,_,_,ID)
	if d.var.chan[ID] then f.openinvite(msg,sender) end
end

function e.CHAT_MSG_EMOTE(msg,sender)
	if d.var.chan.emote then f.openinvite(msg,sender) end
end

function e.CHAT_MSG_SAY(msg,sender)
	if d.var.chan.say then f.openinvite(msg,sender) end
end

function e.CHAT_MSG_SYSTEM(msg)
	if match(msg,d.TR) then d.TradeIni=true end
end

function e.CHAT_MSG_WHISPER_INFORM(_,target)
	f.timer('player','whisper',target)
end

function e.CHAT_MSG_YELL(msg,sender)
	if d.var.chan.yell then f.openinvite(msg,sender) end
end

function e.CVAR_UPDATE(cvar,value)
	if cvar=='BLOCK_TRADES' then
		if value=='0' and not d.var.enable.trade then
			d.var.enable.trade=true
			f.print(L.X_BLOCK_TRADES_OFF)
		elseif value=='1' and d.var.enable.trade then
			d.var.enable.trade=false
			f.print(L.X_BLOCK_TRADES_ON)
		end
		f.register() f.blizzardoptions()
		d['enabletrade']:SetChecked(d.var.enable.trade)
	elseif cvar=='BLOCK_CHAT_CHANNEL_INVITE' then
		if value=='0' and not d.var.enable.chan then
			d.var.enable.chan=true
			f.print(L.X_BLOCK_CHANNELS_OFF)
		elseif value=='1' and d.var.enable.chan then
			d.var.enable.chan=false
			f.print(L.X_BLOCK_CHANNELS_ON)
		end
		f.register() f.blizzardoptions()
		d['enablechan']:SetChecked(d.var.enable.chan)
	end
end

function e.DISABLE_DECLINE_GUILD_INVITE()
	if d.var.enable.guild then return end
	d.var.enable.guild=true
	f.print(L.X_BLOCK_GUILDS_OFF)
	f.register() f.blizzardoptions()
	d['enableguild']:SetChecked(d.var.enable.guild)
end

function e.DUEL_REQUESTED(...)
	f.debug('DUEL',...)
	local name=...
	if not f.allowed(name,'duel') then
		StaticPopup_Hide('DUEL_REQUESTED')
		CancelDuel()
		f.print(L.DECLINED_DUEL,'duel',name)
	end
end

function e.ENABLE_DECLINE_GUILD_INVITE()
	if not d.var.enable.guild then return end
	d.var.enable.guild=false
	f.print(L.X_BLOCK_GUILDS_ON)
	f.register() f.blizzardoptions()
	d['enableguild']:SetChecked(d.var.enable.guild)
end

function e.FRIENDLIST_UPDATE()
	if d.nfriend~=select(2,GetNumFriends()) then
		local count
		count,d.nfriend=GetNumFriends()
		wipe(d.friend)
		for i=1,count do
			local name,_,_,_,online=GetFriendInfo(i)
			if name and online then
				d.friend[(match(name,'%-') and name or name.."-"..d.realm)]=true
			end
		end
	end
end

function e.GROUP_ROSTER_UPDATE()
	if d.OpenInv then d.OpenInv=false f.debug('OPENINV','closed') end
	if d.ngroup~=GetNumGroupMembers() then
		d.ngroup=GetNumGroupMembers()
		wipe(d.group)
		local num,which=4,'party'
		if IsInRaid() then num,which=40,'raid' end
		for i=1,num do
			local name,realm=UnitName(which..i)
			if name then realm=realm or d.realm
			d.group[name.."-"..realm]=true end
			if #d.group==d.ngroup then return end
		end
	end
end

function e.GUILD_INVITE_REQUEST(...)
	f.debug('GUILD',...)
	local name,guild=...
	if not f.allowed(name,'guild') then
		local gi=CreateFrame('frame')
		gi.sfx=GetCVar('Sound_EnableSFX')
		SetCVar('Sound_EnableSFX','0')
		StaticPopupSpecial_Hide(GuildInviteFrame)
		gi:SetScript('OnUpdate',function()
			SetCVar('Sound_EnableSFX',gi.sfx)
			gi:SetScript('OnUpdate',nil)
		end)
		DeclineGuild()
		f.print(L.DECLINED_GUILD,'guild',name,guild)
	end
end

function e.GUILD_ROSTER_UPDATE()
	if d.nguild~=select(2,GetNumGuildMembers()) then
		local count
		count,d.nguild=GetNumGuildMembers()
		wipe(d.guild)
		for i=1,count do
			local name,_,_,_,_,_,_,_,online=GetGuildRosterInfo(i)
			if name and online then
				d.guild[(match(name,'%-') and name or name.."-"..d.realm)]=true
			end
		end
	end
end

function e.PARTY_INVITE_REQUEST(...)
	f.debug('GROUP',...)
	local name,_,_,_,xrealm=...
	if d.OpenInv then d.block=false f.debug(name..': allowed - open invites') return end
	local which=xrealm and 'PARTY_INVITE_XREALM' or 'PARTY_INVITE'
	if not f.allowed(name,'group') then
		StaticPopup_Hide(which)
		LFGInvitePopup:Hide()
		DeclineGroup()
		f.print(L.DECLINED_GROUP,'group',name)
	end
end

function e.PET_BATTLE_PVP_DUEL_REQUESTED(...)
	f.debug('PET',...)
	local name=...
	if not f.allowed(name,'pet') then
		StaticPopup_Hide('PET_BATTLE_PVP_DUEL_REQUESTED')
		CancelPetDuel()
		f.print(L.DECLINED_PET,'pet',name)
	end
end

function e.PETITION_SHOW(...)
	local ptype,_,_,_,name,originator=GetPetitionInfo()
	if originator then return end
	f.debug('CHART',name,...)
	if not f.allowed(name,'chart') then
		HideUIPanel(PetitionFrame)
		ClosePetition()
		if ptype=='guild' then f.print(L.DECLINED_CHART_GUILD,'chart',name)
		else f.print(L.DECLINED_CHART_NIL,'chart',name) end
	end
end

function e.TRADE_CLOSED()
	d.TradeIni=false
end

function e.TRADE_SHOW(...)
	if d.TradeIni then d.block=false return end
	 -- thank you to http://wowprogramming.com/docs/api_types#unitID for that vague Trade UI note next to "npc"
	local name,realm=UnitName('npc')
	name=realm and name.."-"..realm or name.."-"..d.realm
	f.debug('TRADE',name,...)
	if not f.allowed(name,'trade') then
		HideUIPanel(TradeFrame)
		CloseTrade()
		f.print(L.DECLINED_TRADE,'trade',name)
	end
end

function e.UI_INFO_MESSAGE(_,msg)
	if d.UIinfo and d.UI[msg] then UIErrorsFrame:Clear() end d.UIinfo=false
end

function e.ZONE_CHANGED_NEW_AREA()
	if d.OpenInv then d.OpenInv=false f.debug('OPENINV','closed') end
end

 -- calls an event function when that event triggers

DCL:SetScript('OnEvent',function(_,event,...)
	if d.EV[event] then d.block=true end e[event](...)
end)

 -- filters interface sounds

function f.silence(name)
	if d.SO[name] then
		local _,handle=PlaySoundKitID(64,'Master',false)
		if handle then
			StopSound(handle) StopSound(handle-1)
			local si=CreateFrame('frame')
			si:SetScript('OnUpdate',function()
				if not d.block then PlaySoundKitID(d.SO[name]) end
				si:SetScript('OnUpdate',nil)
			end)
		end
	end
end

 -- filters associated system messages if that action is declined

function f.system(_,_,msg)
	if not d.var.gen.disable then
		local msg1=gsub(msg,'|H.-|h%[(.-)%]|h','%1')
		for k,v in next,d.SY do if match(msg1,v) and not f.allowed(match(msg1,v),match(k,'(%a*)'),nil,true) then
			if k=='group1' then f.print(L.DECLINED_GROUP,'group',match(msg1,v)) end
			return true
		end end
	end
end

 -- disables the block trades, block guild invites, and block channel invites options
 -- while editing their display text to reflect the presence of Decliner
 -- Decliner won't be able to let authorized players through if these options are enabled

function f.blizzardoptions()
	if d.var.gen.disable or not d.var.enable.trade then
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetFontObject('GameFontHighlightLeft')
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetText(BLOCK_TRADES)
	else
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetFontObject('GameFontDisableLeft')
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetText(BLOCK_TRADES.." ("..a..")")
		SetCVar('blockTrades','0')
	end
	if d.var.gen.disable or not d.var.enable.guild then
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetFontObject('GameFontHighlightLeft')
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetText(BLOCK_GUILD_INVITES)
	else
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetFontObject('GameFontDisableLeft')
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetText(BLOCK_GUILD_INVITES.." ("..a..")")
		SetAutoDeclineGuildInvites(false)
	end
	if d.var.gen.disable or not d.var.enable.chan then
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetFontObject('GameFontHighlightLeft')
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetText(BLOCK_CHAT_CHANNEL_INVITE)
	else
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetFontObject('GameFontDisableLeft')
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetText(BLOCK_CHAT_CHANNEL_INVITE.." ("..a..")")
		SetCVar('blockChannelInvites','0')
	end
end

 -- event registration

function f.register()
	if d.var.gen.disable then
		DCL:UnregisterEvent('CVAR_UPDATE')
		DCL:UnregisterEvent('DISABLE_DECLINE_GUILD_INVITE')
		DCL:UnregisterEvent('ENABLE_DECLINE_GUILD_INVITE')
		d.header01:SetFontObject('GameFontDisableLarge')
		for i=1,21 do d['text'..format('%02i',i)]:SetFontObject('GameFontDisableSmall') end
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1~='chan' and k1~='open' and k2~='disable' then d[k1..k2]:Disable() end
		end end
		_G[a..'Button50'..'Text']:SetFontObject('GameFontDisableSmall')
		_G[a..'Button51'..'Text']:SetFontObject('GameFontDisableSmall')
		_G[a..'Button52'..'Text']:SetFontObject('GameFontNormalLarge')
	else
		DCL:RegisterEvent('CVAR_UPDATE')
		DCL:RegisterEvent('DISABLE_DECLINE_GUILD_INVITE')
		DCL:RegisterEvent('ENABLE_DECLINE_GUILD_INVITE')
		d.header01:SetFontObject('GameFontNormalLarge')
		for i=1,21 do d['text'..format('%02i',i)]:SetFontObject('GameFontHighlightSmall') end
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1~='chan' and k1~='open' then d[k1..k2]:Enable() end
			if k1~='enable' and d.var.enable[k2]==false then d[k1..k2]:Disable() end
		end end
		_G[a..'Button50'..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[a..'Button51'..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[a..'Button52'..'Text']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.chan then
		DCL:UnregisterEvent('CHANNEL_INVITE_REQUEST')
		d['text01']:SetFontObject('GameFontDisableSmall')
		d['text02']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('CHANNEL_INVITE_REQUEST')
		d['text01']:SetFontObject('GameFontHighlightSmall')
		d['text02']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.duel then
		DCL:UnregisterEvent('DUEL_REQUESTED')
		d['text03']:SetFontObject('GameFontDisableSmall')
		d['text04']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('DUEL_REQUESTED')
		d['text03']:SetFontObject('GameFontHighlightSmall')
		d['text04']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.pet then
		DCL:UnregisterEvent('PET_BATTLE_PVP_DUEL_REQUESTED')
		d['text05']:SetFontObject('GameFontDisableSmall')
		d['text06']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('PET_BATTLE_PVP_DUEL_REQUESTED')
		d['text05']:SetFontObject('GameFontHighlightSmall')
		d['text06']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.group then
		DCL:UnregisterEvent('PARTY_INVITE_REQUEST')
		d['text07']:SetFontObject('GameFontDisableSmall')
		d['text08']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('PARTY_INVITE_REQUEST')
		d['text07']:SetFontObject('GameFontHighlightSmall')
		d['text08']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.guild then
		DCL:UnregisterEvent('GUILD_INVITE_REQUEST')
		d['text09']:SetFontObject('GameFontDisableSmall')
		d['text10']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('GUILD_INVITE_REQUEST')
		d['text09']:SetFontObject('GameFontHighlightSmall')
		d['text10']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.chart then
		DCL:UnregisterEvent('PETITION_SHOW')
		d['text11']:SetFontObject('GameFontDisableSmall')
		d['text12']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('PETITION_SHOW')
		d['text11']:SetFontObject('GameFontHighlightSmall')
		d['text12']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.trade then
		DCL:UnregisterEvent('CHAT_MSG_SYSTEM')
		DCL:UnregisterEvent('TRADE_SHOW')
		DCL:UnregisterEvent('TRADE_CLOSED')
		d['text13']:SetFontObject('GameFontDisableSmall')
		d['text14']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('CHAT_MSG_SYSTEM')
		DCL:RegisterEvent('TRADE_SHOW')
		DCL:RegisterEvent('TRADE_CLOSED')
		d['text13']:SetFontObject('GameFontHighlightSmall')
		d['text14']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not (d.var.enable.duel or d.var.enable.pet or d.var.enable.trade) then
		DCL:UnregisterEvent('UI_INFO_MESSAGE')
	else
		DCL:RegisterEvent('UI_INFO_MESSAGE')
	end
	if d.var.gen.disable or d.var.gen.msgs then
		d.text21:SetFontObject('GameFontDisableSmall')
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1=='msg' then d[k1..k2]:Disable() end
		end end
	else
		d.text21:SetFontObject('GameFontHighlightSmall')
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1=='msg' then d[k1..k2]:Enable() end
			if k1~='enable' and d.var.enable[k2]==false then d[k1..k2]:Disable() end
		end end
	end
	if d.var.gen.disable or d.var.gen.openinv then
		d.OpenInv=false
		d.OIlist:Disable()
		d.OIinput:Disable()
		d.OIbutton:Disable()
		DisableDropDown(d.OIdrop)
		DCL:UnregisterEvent('CHAT_MSG_CHANNEL')
		DCL:UnregisterEvent('CHAT_MSG_EMOTE')
		DCL:UnregisterEvent('CHAT_MSG_SAY')
		DCL:UnregisterEvent('CHAT_MSG_YELL')
		DCL:UnregisterEvent('GROUP_ROSTER_UPDATE')
		DCL:UnregisterEvent('ZONE_CHANGED_NEW_AREA')
		d.OIlist:SetFontObject('GameFontDisableSmall')
		d.OIinput:SetFontObject('GameFontDisableSmall')
	else
		d.OpenInv=d.OItrack and d.OItrack>0
		d.OIlist:Enable()
		d.OIinput:Enable()
		d.OIbutton:Enable()
		EnableDropDown(d.OIdrop)
		DCL:RegisterEvent('CHAT_MSG_CHANNEL')
		DCL:RegisterEvent('CHAT_MSG_EMOTE')
		DCL:RegisterEvent('CHAT_MSG_SAY')
		DCL:RegisterEvent('CHAT_MSG_YELL')
		DCL:RegisterEvent('GROUP_ROSTER_UPDATE')
		DCL:RegisterEvent('ZONE_CHANGED_NEW_AREA')
		d.OIlist:SetFontObject('GameFontHighlightSmall')
		d.OIinput:SetFontObject('GameFontHighlightSmall')
	end
end

 -- Interface Options panel
 -- I would like to thank FlexYourHead of the addon Work_Complete for a ton of comments in their code
 -- It made understanding buttons, text, and the options frame a whole lot easier

function f.create(obj,arg1,arg2,arg3,arg4,arg5,arg6)
	if obj=='b' then
		d.btn=d.btn and d.btn+1 or 1
		local n=a..'Button'..format('%02i',d.btn)
		obj=CreateFrame('CheckButton',n,DCLPanel,'OptionsCheckButtonTemplate')
		obj.tooltipText=arg3
		_G[n..'Text']:SetText(arg5)
		obj:SetCheckedTexture(d.file[arg4])
		obj:SetDisabledCheckedTexture(d.file[arg4])
		obj:SetHitRectInsets(-10,arg6 or -10,-10,-10)
		obj:GetCheckedTexture():SetVertexColor(0.8,0.8,0.8)
		obj:SetPoint('CENTER',DCLPanel,'TOPRIGHT',arg1,arg2)
		obj:GetDisabledCheckedTexture():SetVertexColor(0.4,0.4,0.4)
		local supported=obj:GetDisabledCheckedTexture():SetDesaturated(true)
		if not supported then obj:GetDisabledCheckedTexture():SetAlpha(0.4) end
	else
		d.obj=d.obj and d.obj+1 or 1
		local n=a..'Object'..format('%02i',d.obj)
		if obj=='l' then
			obj=DCLPanel:CreateTexture(n)
			obj:SetSize(600,1)
			obj:SetColorTexture(arg3,arg3,arg3)
			obj:SetPoint('LEFT',DCLPanel,'TOPLEFT',arg1,arg2)
		elseif obj=='t' then
			obj=DCLPanel:CreateFontString(n,'ARTWORK','GameFontNormal')
			obj:SetPoint(arg4,DCLPanel,arg5 and 'TOPLEFT' or 'TOPRIGHT',arg1,arg2)
			obj:SetText(arg3)
		end
	end
	return obj
end

function f.setOIlist()
	sort(d.var.open)
	local text
	if d.var.open[1] then for _,v in next,d.var.open do
		text=text and text.."\n"..tostring(v) or tostring(v)
	end end
	d.OIlist:SetText(text or '')
end

function f.options(a)
	if a=='c' then d.var=ct(d.tmp) elseif a=='d' then d.var=ct(d.def) end
	if a=='r' then
		for k1,v in next,d.def do for k2 in next,v do
			if k1~='chan' and k1~='open' then d[k1..k2]:SetChecked(d.var[k1][k2]) end
		end end
		d.groupgroup.tooltipText=r..L.GROUP_GROUP.."|r" d.guildguild.tooltipText=r..L.GUILD_GUILD.."|r"
		f.setOIlist() f.blizzardoptions() if not d.tmp then d.tmp=ct(d.var) end
	elseif a~='d' then Decliner_SV=ct(d.var) d.tmp=nil end
end

function f.panel()
	DCLPanel.name=a
	InterfaceOptions_AddCategory(DCLPanel,a)
	d.header01=f.create('t',16,-16,a,'TOPLEFT',true)
	d.line01=f.create('l',10,-40,.25)
	d.text01=f.create('t',-460,-64,d.S.CHA,'CENTER')
	d.text02=f.create('t',-460,-80,d.S.INV,'CENTER')
	d.text03=f.create('t',-392,-64,d.S.DUE,'CENTER')
	d.text04=f.create('t',-392,-80,d.S.REQ,'CENTER')
	d.text05=f.create('t',-324,-64,d.S.PET,'CENTER')
	d.text06=f.create('t',-324,-80,d.S.REQ,'CENTER')
	d.text07=f.create('t',-256,-64,d.S.GRO,'CENTER')
	d.text08=f.create('t',-256,-80,d.S.INV,'CENTER')
	d.text09=f.create('t',-188,-64,d.S.GUI,'CENTER')
	d.text10=f.create('t',-188,-80,d.S.INV,'CENTER')
	d.text11=f.create('t',-120,-64,d.S.GUI,'CENTER')
	d.text12=f.create('t',-120,-80,d.S.PTI,'CENTER')
	d.text13=f.create('t',-52,-64,d.S.TRA,'CENTER')
	d.text14=f.create('t',-52,-80,d.S.REQ,'CENTER')
	d.text15=f.create('t',-492,-112,d.S.ENA,'RIGHT')
	d.text16=f.create('t',-492,-160,d.S.WHI,'RIGHT')
	d.text17=f.create('t',-492,-208,d.S.GRM,'RIGHT')
	d.text18=f.create('t',-492,-256,d.S.GUM,'RIGHT')
	d.text19=f.create('t',-492,-304,d.S.REF,'RIGHT')
	d.text20=f.create('t',-492,-352,d.S.BAT,'RIGHT')
	d.text21=f.create('t',-492,-400,d.S.CAL,'RIGHT')
	d.enablechan=f.create('b',-460,-112,g..L.ENABLE_CHAN.."|r",'ch')
	d.enableduel=f.create('b',-392,-112,g..L.ENABLE_DUEL.."|r",'ch')
	d.enablepet=f.create('b',-324,-112,g..L.ENABLE_PET.."|r",'ch')
	d.enablegroup=f.create('b',-256,-112,g..L.ENABLE_GROUP.."|r",'ch')
	d.enableguild=f.create('b',-188,-112,g..L.ENABLE_GUILD.."|r",'ch')
	d.enablechart=f.create('b',-120,-112,g..L.ENABLE_CHART.."|r",'ch')
	d.enabletrade=f.create('b',-52,-112,g..L.ENABLE_TRADE.."|r",'ch')
	d.tellchan=f.create('b',-460,-160,r..L.TELL_CHAN.."|r",'cr')
	d.tellduel=f.create('b',-392,-160,r..L.TELL_DUEL.."|r",'cr')
	d.tellpet=f.create('b',-324,-160,r..L.TELL_PET.."|r",'cr')
	d.tellgroup=f.create('b',-256,-160,r..L.TELL_GROUP.."|r",'cr')
	d.tellguild=f.create('b',-188,-160,r..L.TELL_GUILD.."|r",'cr')
	d.tellchart=f.create('b',-120,-160,r..L.TELL_CHART.."|r",'cr')
	d.telltrade=f.create('b',-52,-160,r..L.TELL_TRADE.."|r",'cr')
	d.groupchan=f.create('b',-460,-208,r..L.GROUP_CHAN.."|r",'cr')
	d.groupduel=f.create('b',-392,-208,r..L.GROUP_DUEL.."|r",'cr')
	d.grouppet=f.create('b',-324,-208,r..L.GROUP_PET.."|r",'cr')
	d.groupgroup=f.create('b',-256,-208,r..L.GROUP_GROUP.."|r",'cr')
	d.groupguild=f.create('b',-188,-208,r..L.GROUP_GUILD.."|r",'cr')
	d.groupchart=f.create('b',-120,-208,r..L.GROUP_CHART.."|r",'cr')
	d.grouptrade=f.create('b',-52,-208,r..L.GROUP_TRADE.."|r",'cr')
	d.guildchan=f.create('b',-460,-256,r..L.GUILD_CHAN.."|r",'cr')
	d.guildduel=f.create('b',-392,-256,r..L.GUILD_DUEL.."|r",'cr')
	d.guildpet=f.create('b',-324,-256,r..L.GUILD_PET.."|r",'cr')
	d.guildgroup=f.create('b',-256,-256,r..L.GUILD_GROUP.."|r",'cr')
	d.guildguild=f.create('b',-188,-256,r..L.GUILD_GUILD.."|r",'cr')
	d.guildchart=f.create('b',-120,-256,r..L.GUILD_CHART.."|r",'cr')
	d.guildtrade=f.create('b',-52,-256,r..L.GUILD_TRADE.."|r",'cr')
	d.friendchan=f.create('b',-460,-304,r..L.FRIEND_CHAN.."|r",'cr')
	d.friendduel=f.create('b',-392,-304,r..L.FRIEND_DUEL.."|r",'cr')
	d.friendpet=f.create('b',-324,-304,r..L.FRIEND_PET.."|r",'cr')
	d.friendgroup=f.create('b',-256,-304,r..L.FRIEND_GROUP.."|r",'cr')
	d.friendguild=f.create('b',-188,-304,r..L.FRIEND_GUILD.."|r",'cr')
	d.friendchart=f.create('b',-120,-304,r..L.FRIEND_CHART.."|r",'cr')
	d.friendtrade=f.create('b',-52,-304,r..L.FRIEND_TRADE.."|r",'cr')
	d.bnetchan=f.create('b',-460,-352,r..L.BNET_CHAN.."|r",'cr')
	d.bnetduel=f.create('b',-392,-352,r..L.BNET_DUEL.."|r",'cr')
	d.bnetpet=f.create('b',-324,-352,r..L.BNET_PET.."|r",'cr')
	d.bnetgroup=f.create('b',-256,-352,r..L.BNET_GROUP.."|r",'cr')
	d.bnetguild=f.create('b',-188,-352,r..L.BNET_GUILD.."|r",'cr')
	d.bnetchart=f.create('b',-120,-352,r..L.BNET_CHART.."|r",'cr')
	d.bnettrade=f.create('b',-52,-352,r..L.BNET_TRADE.."|r",'cr')
	d.msgchan=f.create('b',-460,-400,r..L.MSG_CHAN.."|r",'cr')
	d.msgduel=f.create('b',-392,-400,r..L.MSG_DUEL.."|r",'cr')
	d.msgpet=f.create('b',-324,-400,r..L.MSG_PET.."|r",'cr')
	d.msggroup=f.create('b',-256,-400,r..L.MSG_GROUP.."|r",'cr')
	d.msgguild=f.create('b',-188,-400,r..L.MSG_GUILD.."|r",'cr')
	d.msgchart=f.create('b',-120,-400,r..L.MSG_CHART.."|r",'cr')
	d.msgtrade=f.create('b',-52,-400,r..L.MSG_TRADE.."|r",'cr')
	d.line02=f.create('l',10,-428,.25)
	d.genopeninv=f.create('b',-256,-456,r..L.X_DISABLE_OPENINV.."|r",'cr',d.S.DIO..d.S.INV,-200)
	d.genmsgs=f.create('b',-256,-496,r..L.X_DISABLE_MESSAGES.."|r",'cr',d.S.DIS..d.S.CAL,-200)
	d.gendisable=f.create('b',-256,-536,r..L.X_DISABLE_DECLINER.."|r",'cr',d.S.DIS..a,-200)

	for k1,v1 in next,d.def do for k2 in next,v1 do
		if k1==k2 then
			d[k1..k2]:SetScript('PostClick',function(self)
				if d.stp[k2] then StopSound(d.stp[k2]) end
				self:SetChecked(true) d.var[k1][k2]=false
				local x=random(1,55555)
				self.tooltipText=r..'ID: '..x..'|r'
				d.ply[k2],d.stp[k2]=PlaySoundKitID(x)
				if d.ply[k2] then self:SetCheckedTexture(d.file.ha)
				else self:SetCheckedTexture(d.file.sa) end
			end)
		elseif k1~='chan' and k1~='open' then
			d[k1..k2]:SetScript('PostClick',function(self)
				d.var[k1][k2]=self:GetChecked()
				f.register() f.blizzardoptions()
			end)
		end
	end end

	-- the following inputbox, button, and scrolling list inspired by funkydude's BadBoy_CCleaner chat filter options

	d.OIinput=CreateFrame('EditBox',a..'OpenInvInput',DCLPanel,'InputBoxTemplate')
	d.OIinput:SetPoint('TOPRIGHT',-300,-461)
	d.OIinput:SetAutoFocus(false)
	d.OIinput:EnableMouse(true)
	d.OIinput:SetHeight(20)
	d.OIinput:SetWidth(112)
	d.OIinput:SetScript('OnEnterPressed',function()d.OIbutton:Click()end)
	d.OIinput:SetScript('OnEscapePressed',function(frame)
		frame:SetText('')
		frame:ClearFocus()
	end)
	d.OIinput:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT')
			GameTooltip:SetText(L.X_OPENINV_INPUT,nil,nil,nil,nil,true)
		end
	end)
	d.OIinput:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	d.OIbutton=CreateFrame('Button',a..'OpenInvButton',DCLPanel,'UIPanelButtonTemplate')
	d.OIbutton:SetPoint('TOPRIGHT',d.OIinput,'BOTTOMRIGHT',1,0)
	d.OIbutton:SetPoint('TOPLEFT',d.OIinput,'BOTTOMLEFT',-6,0)
	d.OIbutton:SetNormalFontObject('GameFontNormalSmall')
	d.OIbutton:SetHighlightFontObject('GameFontHighlightSmall')
	d.OIbutton:SetDisabledFontObject('GameFontDisableSmall')
	d.OIbutton:SetText(d.S.ADD)
	d.OIbutton:SetScript('OnClick',function()
		d.OIinput:ClearFocus()
		local t=d.OIinput:GetText()
		if t:find('^%s*$') then d.OIinput:SetText('') return end
		local found
		for k,v in next,d.var.open do if t==v then found=k end end
		if found then remove(d.var.open,found) else insert(d.var.open,t) end
		f.setOIlist()
		d.OIinput:SetText('')
	end)
	d.OIbutton:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT',-1)
			GameTooltip:SetText(L.X_OPENINV_BUTTON,nil,nil,nil,nil,true)
		end
	end)
	d.OIbutton:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	d.OIdrop=CreateFrame('Frame',a..'OpenInvDropDown',DCLPanel,'UIDropDownMenuTemplate')
	d.OIdrop:SetPoint('TOPLEFT',d.OIbutton,'BOTTOMLEFT',-16,0)
	UIDropDownMenu_SetWidth(d.OIdrop,100)
	UIDropDownMenu_SetText(d.OIdrop,d.S.CHS)
	UIDropDownMenu_Initialize(d.OIdrop,function(frame)
		local info=UIDropDownMenu_CreateInfo()
		info.func=function(self,arg1)d.var.chan[arg1]=not d.var.chan[arg1]end
		info.text,info.checked,info.arg1=d.S.EMO,d.var.chan.emote,'emote'
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=d.S.SAY,d.var.chan.say,'say'
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=d.S.YEL,d.var.chan.yell,'yell'
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=d.S.GEN,d.var.chan[1],1
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=d.S.TRA,d.var.chan[2],2
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=d.S.LOO,d.var.chan[26],26
		UIDropDownMenu_AddButton(info)
	end)
	d.OIdrop:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT',-16)
			GameTooltip:SetText(L.X_OPENINV_MENU,nil,nil,nil,nil,true)
		end
	end)
	d.OIdrop:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	d.OIback=CreateFrame('Frame',a..'OpenInvBackdrop',DCLPanel)
	d.OIback:SetPoint('TOPRIGHT',d.OIinput,'TOPLEFT',-24,0)
	d.OIback:SetPoint('BOTTOMLEFT',d.OIdrop,'BOTTOMLEFT',-119,6)
	d.OIback:SetBackdrop({
		bgFile='Interface\\Tooltips\\UI-Tooltip-Background',
		edgeFile='Interface\\Tooltips\\UI-Tooltip-Border',
		insets={left=2,right=2,top=2,bottom=2},
		tile=true,tileSize=8,edgeSize=8,})
	d.OIback:SetBackdropColor(0,0,0,.75)
	d.OIback:SetBackdropBorderColor(.25,.25,.25)
	d.OIback:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT')
			GameTooltip:SetText(L.X_OPENINV_LIST,nil,nil,nil,nil,true)
		end
	end)
	d.OIback:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	d.OIscroll=CreateFrame('ScrollFrame',a..'OpenInvScroll',DCLPanel,'UIPanelScrollFrameTemplate')
	d.OIscroll:SetPoint('LEFT',d.OIback,'LEFT',4.5,0)
	d.OIscroll:SetPoint('RIGHT',d.OIback,'RIGHT',-4.5,0)
	d.OIscroll:SetPoint('TOP',d.OIback,'TOP',0,-4)
	d.OIscroll:SetPoint('BOTTOM',d.OIback,'BOTTOM',0,4)

	d.OIlist=CreateFrame('EditBox',a..'OpenInvList',DCLPanel)
	d.OIlist:SetMaxLetters(9999)
	d.OIlist:SetAutoFocus(false)
	d.OIlist:EnableMouse(false)
	d.OIlist:SetMultiLine(true)
	d.OIlist:SetWidth(120)
	d.OIlist:Show()

	d.OIscroll:SetScrollChild(d.OIlist)

	DCLPanel.okay=function()f.options()end
	DCLPanel.cancel=function()f.options('c')f.register()end
	DCLPanel.default=function()f.options('d')f.register()end
	DCLPanel.refresh=function()f.options('r')end
	SLASH_DECLINER1='/decliner'
	SLASH_DECLINER2='/decline'
	SLASH_DECLINER3='/dcl'
	SlashCmdList.DECLINER=function()InterfaceOptionsFrame_OpenToCategory(a)end
end

 -- fix Interface Options bug where addons don't open to their own options
hooksecurefunc('InterfaceOptions_AddCategory',InterfaceAddOnsList_Update)
hooksecurefunc('InterfaceOptions_AddCategory',InterfaceCategoryList_Update)